#!/usr/bin/perl

use strict;
use warnings;

use feature qw(state);

use English qw(-no_match_vars);
use File::Temp;
use HTML::Tidy5;
use Test::HTML::Tidy5 qw();
use Test::More;
use Test::Mojo::WithRoles 'SubmitForm';

use AccessCheck::Data::DB;
use AccessCheck::Data::Entity;

plan(skip_all => 'live database required') unless
    $ENV{TEST_DB_HOST} &&
    $ENV{TEST_DB_NAME} &&
    $ENV{TEST_DB_TYPE};

plan tests => 16;

sub named_subtest {
    my ($name, $code, @args) = @_;

    my $tb = Test::More->builder();
    return $tb->subtest($name, $code, @args, $name);
}

sub setup {
    my %args = @_;

    my $name = $args{name};
    my $host = $args{host};
    my $type = $args{type};
    my $username = $args{username} || '';
    my $password = $args{password} || '';

    system("mysqladmin --host=$host --user=$username --password=$password --force drop $name >/dev/null");
    system("mysqladmin --host=$host --user=$username --password=$password create $name >/dev/null") == 0
        or die "can't run mysqladmin: $CHILD_ERROR\n";
    system("mysql --host=$host --user=$username --password=$password $name < conf/manager.sql") == 0
        or die "can't run mysql: $CHILD_ERROR\n";

    AccessCheck::Data::DB->register_db(
        driver   => $args{type},
        host     => $host,
        database => $name,
        password => $password,
        username => $username,
    );
    my $db = AccessCheck::Data::DB->new();

    my $entity = AccessCheck::Data::Entity->new(
        db               => $db,
        type             => 'sp',
        entityid         => 'https://sp.renater.fr/',
        display_name     => 'https://sp.renater.fr/ (Test SP)',
        information_url  => 'https://sp.renater.fr',
        organization_url => 'http://www.renater.fr ',
        contacts         => ['contact1@renater.fr', 'contact2@renater.fr'],
        federations      => ['edugain'],
    );
    $entity->save();

    my $temp_dir = File::Temp->newdir(CLEANUP => $ENV{TEST_DEBUG} ? 0 : 1);
    diag("temp dir: $temp_dir") if $ENV{TEST_DEBUG};

    my $log_file  = sprintf("%s/test.log", $temp_dir);
    my $conf_file = sprintf("%s/test.conf", $temp_dir);

    open(my $handle, '>', $conf_file);
    print {$handle} <<EOF;
[setup]
templates_theme = edugain

[federations]
edugain = t/edugain.xml

[app]
url = https://my.fqdn/accountmanager
support_email = support\@my.fqdn
name = eduGAIN Access Check

[logger]
file = $log_file
level = debug

[mailer]

[idp]

[database]
host     = $host
name     = $name
type     = $type
username = $username
password = $password
EOF
    close($handle);

    $ENV{ACCESS_CHECK_CONFIG} = $conf_file;

    return $temp_dir;
}

sub get_test_object {
    my %args = @_;

    my $app = $args{app} || 'AccessCheck::App';

    my $t = Test::Mojo::WithRoles->new($app);
    my $ua = $t->ua();
    $ua->max_redirects(1);
    $ua->on(start => sub {
        my (undef, $tx) = @_;
        $tx->req()->headers()->from_hash($args{user})
            if $args{user} and !$tx->previous();
        $tx->req()->headers()->accept_language($args{language})
            if $args{language};
    });

    $t->app()->log()->info(sprintf('new test: %s', $args{test}))
        if $args{test};

    return $t;
}

sub html_ok {
    my ($res) = @_;

    state $tidy =  HTML::Tidy5->new({
        'drop-empty-elements' => 0
    });
    $tidy->ignore(type => TIDY_INFO);
    $tidy->ignore(text => qr/<img> lacks "alt" attribute/);
    local $ENV{LANG} = 'C';
    Test::HTML::Tidy5::html_tidy_ok($tidy, $res->text(), "HTML content OK");
}

sub diag_file {
    my ($res, $dir) = @_;

    return if !$ENV{TEST_DEBUG};

    state $count = 1;
    my $file = sprintf("%s/test_%i.html", $dir, $count);
    diag("content saved as $file");

    open (my $handle, '>', $file) or die($ERRNO);
    print $handle $res->text();
    close $handle;

    $count++;
}

my $test_dir = setup(
    host     => $ENV{TEST_DB_HOST},
    name     => $ENV{TEST_DB_NAME}, 
    type     => $ENV{TEST_DB_TYPE},
    username => $ENV{TEST_DB_USERNAME},
    password => $ENV{TEST_DB_PASSWORD}
);

named_subtest "index page" => sub {
    my $t = get_test_object(test => $_[0]);

    $t->get_ok('/')
      ->status_is(200)
      ->text_is('html head title' => 'eduGAIN Access Check')
      ->text_is('a[href=/select/entity]'  => 'Get started', 'get started button');

    my $res = $t->tx()->res();
    html_ok($res) or diag_file($res, $test_dir);
};

named_subtest "SP selection page" => sub {
    my $t = get_test_object(test => $_[0]);

    $t->get_ok('/select/entity')
      ->status_is(200)
      ->text_is('html head title' => 'eduGAIN Access Check', 'expected title')
      ->element_exists('select[id=all][name=all]', 'SP selection widget');

    my $res = $t->tx()->res();
    html_ok($res) or diag_file($res, $test_dir);
};

named_subtest "email selection page, missing entityid" => sub {
    my $t = get_test_object(test => $_[0]);

    $t->get_ok('/select/email')
      ->status_is(200)
      ->text_is('html head title' => 'eduGAIN Access Check', 'expected title')
      ->content_like(qr/Error:[\n\s]+missing parameter 'entityid'/, 'expected error message');

    my $res = $t->tx()->res();
    html_ok($res) or diag_file($res, $test_dir);
};

named_subtest "email selection page, invalid entityid" => sub {
    my $t = get_test_object(test => $_[0]);

    $t->get_ok('/select/email' => form => {entityid => 'foo'})
      ->status_is(200)
      ->text_is('html head title' => 'eduGAIN Access Check', 'expected title')
      ->content_like(qr/Error:[\n\s]+invalid parameter 'entityid'/, 'expected error message');

    my $res = $t->tx()->res();
    html_ok($res) or diag_file($res, $test_dir);
};

named_subtest "email selection page, valid entityid" => sub {
    my $t = get_test_object(test => $_[0]);

    $t->get_ok('/select/email' => form => {entityid => 'https://sp.renater.fr/'})
      ->status_is(200)
      ->text_is('html head title' => 'eduGAIN Access Check', 'expected title')
      ->element_exists('input[name=email][value=contact1@renater.fr]', 'email selection widget');

    my $res = $t->tx()->res();
    html_ok($res) or diag_file($res, $test_dir);
};

named_subtest "challenge sending page, missing CSRF token" => sub {
    my $t = get_test_object(test => $_[0]);

    $t->get_ok('/challenge/send')
      ->status_is(403)
      ->text_is('html head title' => 'eduGAIN Access Check', 'expected title')
      ->content_like(qr/Error:[\n\s]+missing CSRF token/, 'expected error message');

    my $res = $t->tx()->res();
    html_ok($res) or diag_file($res, $test_dir);
};

named_subtest "challenge sending page, missing entityid" => sub {
    my $t = get_test_object(test => $_[0]);

    # neutralize CSRF token check, as we short-circuit the form
    $t->app()->hook(before_dispatch => sub { $_[0]->session(csrf_token => 'foo') });

    $t->get_ok('/challenge/send' => form => {token => 'foo'})
      ->status_is(200)
      ->text_is('html head title' => 'eduGAIN Access Check', 'expected title')
      ->content_like(qr/Error:[\n\s]+missing parameter 'entityid'/, 'expected error message');

    my $res = $t->tx()->res();
    html_ok($res) or diag_file($res, $test_dir);
};

named_subtest "challenge sending page, invalid entityid" => sub {
    my $t = get_test_object(test => $_[0]);

    # neutralize CSRF token check, as we short-circuit the form
    $t->app()->hook(before_dispatch => sub { $_[0]->session(csrf_token => 'foo') });

    $t->get_ok('/challenge/send' => form => {token => 'foo', entityid => 'foo'})
      ->status_is(200)
      ->text_is('html head title' => 'eduGAIN Access Check', 'expected title')
      ->content_like(qr/Error:[\n\s]+invalid parameter 'entityid'/, 'expected error message');

    my $res = $t->tx()->res();
    html_ok($res) or diag_file($res, $test_dir);
};

named_subtest "challenge sending page, valid entityid, missing email" => sub {
    my $t = get_test_object(test => $_[0]);

    # neutralize CSRF token check, as we short-circuit the form
    $t->app()->hook(before_dispatch => sub { $_[0]->session(csrf_token => 'foo') });

    $t->get_ok('/challenge/send' => form => {token => 'foo', entityid => 'https://sp.renater.fr/'})
      ->status_is(200)
      ->text_is('html head title' => 'eduGAIN Access Check', 'expected title')
      ->content_like(qr/Error:[\n\s]+missing parameter 'email'/, 'expected error message');

    my $res = $t->tx()->res();
    html_ok($res) or diag_file($res, $test_dir);
};

named_subtest "challenge sending page, valid entityid, invalid email" => sub {
    my $t = get_test_object(test => $_[0]);

    # neutralize CSRF token check, as we short-circuit the form
    $t->app()->hook(before_dispatch => sub { $_[0]->session(csrf_token => 'foo') });

    $t->get_ok('/challenge/send' => form => {token => 'foo', entityid => 'https://sp.renater.fr/', email => 'foo'})
      ->status_is(200)
      ->text_is('html head title' => 'eduGAIN Access Check', 'expected title')
      ->content_like(qr/Error:[\n\s]+invalid parameter 'email'/, 'expected error message');

    my $res = $t->tx()->res();
    html_ok($res) or diag_file($res, $test_dir);
};

named_subtest "challenge validation page, missing entityid" => sub {
    my $t = get_test_object(test => $_[0]);

    $t->get_ok('/challenge/validate')
      ->status_is(200)
      ->text_is('html head title' => 'eduGAIN Access Check', 'expected title')
      ->content_like(qr/Error:[\n\s]+missing parameter 'entityid'/, 'expected error message');

    my $res = $t->tx()->res();
    html_ok($res) or diag_file($res, $test_dir);
};

named_subtest "challenge validation page, invalid entityid" => sub {
    my $t = get_test_object(test => $_[0]);

    $t->get_ok('/challenge/validate' => form => {entityid => 'foo'})
      ->status_is(200)
      ->text_is('html head title' => 'eduGAIN Access Check', 'expected title')
      ->content_like(qr/Error:[\n\s]+invalid parameter 'entityid'/, 'expected error message');

    my $res = $t->tx()->res();
    html_ok($res) or diag_file($res, $test_dir);
};

named_subtest "challenge validation page, valid entityid, missing email" => sub {
    my $t = get_test_object(test => $_[0]);

    $t->get_ok('/challenge/validate' => form => {entityid => 'https://sp.renater.fr/'})
      ->status_is(200)
      ->text_is('html head title' => 'eduGAIN Access Check', 'expected title')
      ->content_like(qr/Error:[\n\s]+missing parameter 'email'/, 'expected error message');

    my $res = $t->tx()->res();
    html_ok($res) or diag_file($res, $test_dir);
};

named_subtest "challenge validation page, valid entityid, invalid email" => sub {
    my $t = get_test_object(test => $_[0]);

    $t->get_ok('/challenge/validate' => form => {entityid => 'https://sp.renater.fr/', email => 'foo'})
      ->status_is(200)
      ->text_is('html head title' => 'eduGAIN Access Check', 'expected title')
      ->content_like(qr/Error:[\n\s]+invalid parameter 'email'/, 'expected error message');

    my $res = $t->tx()->res();
    html_ok($res) or diag_file($res, $test_dir);
};

named_subtest "challenge validation page, valid entityid, valid email" => sub {
    my $t = get_test_object(test => $_[0]);

    $t->get_ok('/challenge/validate' => form => {entityid => 'https://sp.renater.fr/', email => 'contact1@renater.fr'})
      ->status_is(200)
      ->text_is('html head title' => 'eduGAIN Access Check', 'expected title')
      ->text_is('div.content h2'  => 'Complete email challenge', 'expected form title');

    my $res = $t->tx()->res();
    html_ok($res) or diag_file($res, $test_dir);
};

named_subtest "index page, french version" => sub {
    my $t = get_test_object(test => $_[0], language => 'fr-FR,fr;q=0.9');

    $t->get_ok('/')
      ->status_is(200)
      ->text_is('html head title' => 'eduGAIN Access Check')
      ->text_is('html body a[href=/select/entity]'  => 'Commencer', 'get started button');

    my $res = $t->tx()->res();
    html_ok($res) or diag_file($res, $test_dir);
};
